using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Text;
using Microsoft.DirectX.Direct3D;

namespace BenLincoln.TheLostWorlds.CDTextures
{
    public class SoulReaverPCTextureFile : SoulReaverMonolithicTextureFile
    {
        // Methods
        public SoulReaverPCTextureFile(string path)
            : base(path)
        {
            base._FileType = TextureFileType.SoulReaverPC;
            base._FileTypeName = "Soul Reaver (PC) Uncompressed ARGB 1555 Monolithic Texture File";
            base._HeaderLength = 0x1000;
            base._TextureLength = 0x20000;
            base._TextureCount = this._GetTextureCount();
        }

        protected override Texture _GetTexture(Device device, int index)
        {
            return Texture.FromBitmap(device, this._GetTextureAsBitmap(index), Usage.None, Pool.Managed);
        }

        protected override Bitmap _GetTextureAsBitmap(int index)
        {
            int num = base._HeaderLength + (index * base._TextureLength);
            if (num > (base._FileInfo.Length - base._TextureLength))
            {
                throw new TextureFileException("An index was specified which resulted in an offset greater than the maximum allowed value.");
            }
            int num9 = 3;
            int num10 = 3;
            int num11 = 3;
            Color color = new Color();
            Bitmap bitmap = new Bitmap(0x100, 0x100);
            try
            {
                FileStream input = new FileStream(base._FilePath, FileMode.Open, FileAccess.Read);
                BinaryReader reader = new BinaryReader(input);
                input.Seek((long)num, SeekOrigin.Begin);
                for (ushort i = 0; i <= 0xff; i = (ushort)(i + 1))
                {
                    for (ushort j = 0; j <= 0xff; j = (ushort)(j + 1))
                    {
                        ushort num8 = reader.ReadUInt16();
                        ushort alpha = num8;
                        ushort red = num8;
                        ushort green = num8;
                        ushort blue = num8;
                        alpha = (ushort)(alpha >> 15);
                        red = (ushort)(red << 1);
                        red = (ushort)(red >> 11);
                        green = (ushort)(green << 6);
                        green = (ushort)(green >> 11);
                        blue = (ushort)(blue << 11);
                        blue = (ushort)(blue >> 11);
                        if (alpha > 0)
                        {
                            alpha = 0xff;
                        }
                        red = (ushort)(red << num9);
                        green = (ushort)(green << num10);
                        blue = (ushort)(blue << num11);
                        color = Color.FromArgb(alpha, red, green, blue);
                        bitmap.SetPixel(j, i, color);
                    }
                }
                reader.Close();
                input.Close();
            }
            catch (Exception exception)
            {
                throw new TextureFileException("Error reading the specified texture.", exception);
            }
            return bitmap;
        }

        public override MemoryStream GetDataAsStream(int index)
        {
            throw new Exception("The method or operation is not implemented.");
        }
    }
}
