using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Reflection;
using System.Text;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using DX = Microsoft.DirectX;
using D3DX = Microsoft.DirectX.Direct3D;
using Imaging = System.Drawing.Imaging;

namespace ModelEx3
{
    #region Enumerations

    public enum ModelFormat
    {
        Binary = D3DX.XFileFormat.Binary,
        Text = D3DX.XFileFormat.Text,
        BinaryCompressed = D3DX.XFileFormat.Binary | D3DX.XFileFormat.Compressed,
        TextCompressed = D3DX.XFileFormat.Text | D3DX.XFileFormat.Compressed,
        LastUsed = 4
    }

    public enum ImageFormat
    {
        Bmp = 0,
        Jpeg = 1,
        Png = 2,
        LastUsed = 3
    }

    public enum CullType
    {
        None = D3DX.Cull.None,
        Clockwise = D3DX.Cull.Clockwise,
        CounterClockwise = D3DX.Cull.CounterClockwise
    }

    public enum RenderFlags
    {
        None = 0x00,
        Lighting = 0x01,
        Textures = 0x02,
        TextureFiltering = 0x04,
        WireFrame = 0x08,
        Scaling = 0x10,
    }

    #endregion
    
    public interface ModelExList<T> : IEnumerable<T>
    {
        int Count { get; }
        T this[int iIndex] { get; }
    }

    partial class ModelExSystem
    {
        protected Device m_xDevice;
        protected ModelExFileList<ModelExEffectFile> m_xEffectFiles;
        protected ModelExFileList<ModelExTextureFile> m_xTextureFiles;
        protected ModelExFileList<ModelExModelFile> m_xModelFiles;
        protected ModelExFileList<ModelExObjectFile> m_xObjectFiles;

        public Device Device { get { return m_xDevice; } }
        public ModelExEffectFile[] EffectFiles { get { return m_xEffectFiles.ToArray(); } }
        public ModelExTextureFile[] TextureFiles { get { return m_xTextureFiles.ToArray(); } }
        public ModelExModelFile[] ModelFiles { get { return m_xModelFiles.ToArray(); } }
        public ModelExObjectFile[] ObjectFiles { get { return m_xObjectFiles.ToArray(); } }

        public ModelExSystem(ref Device xDevice)
        {
            m_xDevice = xDevice;
            m_xEffectFiles = new ModelExFileList<ModelExEffectFile>();
            m_xTextureFiles = new ModelExFileList<ModelExTextureFile>();
            m_xModelFiles = new ModelExFileList<ModelExModelFile>();
            m_xObjectFiles = new ModelExFileList<ModelExObjectFile>();
        }

        public void AddEffectFile(ModelExEffectFile xEffectFile)
        {
            m_xEffectFiles.Add(xEffectFile);
        }

        public ModelExEffectFile FindEffectFile(String strEffectName)
        {
            return m_xEffectFiles.Find(strEffectName);
        }

        public void ClearEffectFiles()
        {
            m_xEffectFiles.Clear();
        }

        public void AddTextureFile(ModelExTextureFile xTextureFile)
        {
            m_xTextureFiles.Add(xTextureFile);
        }

        public ModelExTextureFile FindTextureFile(String strTextureName)
        {
            return m_xTextureFiles.Find(strTextureName);
        }

        public void ClearTextureFiles()
        {
            m_xTextureFiles.Clear();
        }

        public void AddModelFile(ModelExModelFile xModelFile)
        {
            m_xModelFiles.Add(xModelFile);
        }

        public ModelExModelFile FindModelFile(String strModelName)
        {
            return m_xModelFiles.Find(strModelName);
        }

        public void ClearModelFiles()
        {
            m_xModelFiles.Clear();
        }

        public void AddObjectFile(ModelExObjectFile xObjectFile)
        {
            m_xObjectFiles.Add(xObjectFile);
        }

        public ModelExObjectFile FindObjectFile(String strObjectName)
        {
            return m_xObjectFiles.Find(strObjectName);
        }

        public void ClearObjectFiles()
        {
            m_xObjectFiles.Clear();
        }
    }
}
