using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.IO;
using System.Text;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace ModelEx3
{
    class ModelExView : DirectXView
    {
        #region Enumerations

        public enum ControlType
        {
            None = 0,
            OrbitCamera = 1,
            FlyCamera = 2,
            MapCamera = 3,
            RotateObject = 4,
        }

        #endregion

        #region Members

        ModelExSystem m_xSystem;
        ModelExObjectInstance m_xObject;
        //ModelExObjectInstance m_xCharacter;
        RenderFlags m_eRenderFlags;
        String m_strLastError;

        // Not gonna work, the offsets are in bytes :(
        //[System.Runtime.InteropServices.StructLayout(System.Runtime.InteropServices.LayoutKind.Explicit)]
        //struct RenderFlagStruct
        //{
        //    [System.Runtime.InteropServices.FieldOffset(0)]
        //    public RenderFlags m_bRenderFlags;

        //    [System.Runtime.InteropServices.FieldOffset(0)]
        //    public bool Lighting;

        //    [System.Runtime.InteropServices.FieldOffset(1)]
        //    public bool Textures;

        //    [System.Runtime.InteropServices.FieldOffset(2)]
        //    public bool TextureFiltering;

        //    [System.Runtime.InteropServices.FieldOffset(3)]
        //    public bool WireFrame;
        //}

        CullType m_eCullMode;

        ModelExCamera m_xCamera;
        ModelExArcBall m_xArcBall;
        Point m_xLastMousePosition;
        Point m_xLastMouseDownPosition;
        Control m_xTextOutputControl;
        ControlType m_eControlMode;

        #endregion

        #region Constructor

        public ModelExView() : base()
        {
            m_xBackGroundColour = Color.Gray;
            m_xLastMousePosition = new Point();
            MouseUp += new MouseEventHandler(ModelExView_MouseUp);
            MouseDown += new MouseEventHandler(ModelExView_MouseDown);
            MouseMove += new MouseEventHandler(ModelExView_MouseMove);
            MouseWheel += new MouseEventHandler(ModelExView_MouseWheel);
            PreviewKeyDown += new PreviewKeyDownEventHandler(ModelExView_PreviewKeyDown);
            GotFocus += new EventHandler(ModelExView_ChangeFocus);
            LostFocus += new EventHandler(ModelExView_ChangeFocus);
            UpdateTextOutputControl();
        }

        #endregion

        #region Event Handlers

        void ModelExView_ChangeFocus(object sender, EventArgs e)
        {
            UpdateTextOutputControl();
        }

        void ModelExView_PreviewKeyDown(object sender, PreviewKeyDownEventArgs e)
        {
            if (m_xArcBall != null && m_xCamera != null)
            {
                #region Dragging not allowed

                if (m_eControlMode == ControlType.OrbitCamera ||
                    m_eControlMode == ControlType.FlyCamera)
                {

                    if (!m_xArcBall.IsMoving)
                    {
                        switch (e.KeyCode)
                        {
                            case Keys.Up:
                                m_xCamera.RotateY(2.0f * ((float)Math.PI) / -200.0f);
                                break;
                            case Keys.Down:
                                m_xCamera.RotateY(2.0f * ((float)Math.PI) / 200.0f);
                                break;
                            case Keys.Left:
                                {
                                    if (e.Control)
                                    {
                                        m_xCamera.RotateZ(2.0f * ((float)Math.PI) / 200.0f);
                                    }
                                    else
                                    {
                                        m_xCamera.RotateX(2.0f * ((float)Math.PI) / -200.0f);
                                    }
                                    break;
                                }
                            case Keys.Right:
                                {
                                    if (e.Control)
                                    {
                                        m_xCamera.RotateZ(2.0f * ((float)Math.PI) / -200.0f);
                                    }
                                    else
                                    {
                                        m_xCamera.RotateX(2.0f * ((float)Math.PI) / 200.0f);
                                    }
                                    break;
                                }
                        }
                    }
                }

                #endregion

                #region Dragging allowed

                if (m_eControlMode == ControlType.OrbitCamera ||
                    m_eControlMode == ControlType.FlyCamera ||
                    m_eControlMode == ControlType.MapCamera)
                {
                    switch (e.KeyCode)
                    {
                        case Keys.W:
                        {
                            if (m_xCamera.Mode == ModelExCamera.CameraMode.Fly)
                            {
                                m_xCamera.MoveForward(0.01f);
                            }
                            else if (m_xCamera.Mode == ModelExCamera.CameraMode.Map)
                            {
                                m_xCamera.MoveUpward(0.01f);
                            }
                            break;
                        }
                        case Keys.S:
                        {
                            if (m_xCamera.Mode == ModelExCamera.CameraMode.Fly)
                            {
                                m_xCamera.MoveForward(-0.01f);
                            }
                            else if (m_xCamera.Mode == ModelExCamera.CameraMode.Map)
                            {
                                m_xCamera.MoveUpward(-0.01f);
                            }
                            break;
                        }
                        case Keys.A:
                        {
                            if (m_xCamera.Mode == ModelExCamera.CameraMode.Fly ||
                                m_xCamera.Mode == ModelExCamera.CameraMode.Map)
                            {
                                m_xCamera.MoveRight(-0.01f);
                            }
                            break;
                        }
                        case Keys.D:
                        {
                            if (m_xCamera.Mode == ModelExCamera.CameraMode.Fly ||
                                m_xCamera.Mode == ModelExCamera.CameraMode.Map)
                            {
                                m_xCamera.MoveRight(0.01f);
                            }
                            break;
                        }
                        case Keys.U:
                        {
                            if (m_xCamera.Mode == ModelExCamera.CameraMode.Fly ||
                                m_xCamera.Mode == ModelExCamera.CameraMode.Map)
                            {
                                m_xCamera.RemoveRoll();
                            }
                            break;
                        }
                    }
                }

                #endregion
            }
        }

        void ModelExView_MouseWheel(object sender, MouseEventArgs e)
        {
        }

        void ModelExView_MouseUp(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                if (m_xArcBall != null)
                {
                    m_xArcBall.EndMove();
                }
            }
        }

        void ModelExView_MouseDown(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                if (m_xArcBall != null)
                {
                    m_xArcBall.BeginMove();
                }
                m_xLastMouseDownPosition = new Point(e.X, e.Y);
            }
        }

        void ModelExView_MouseMove(object sender, MouseEventArgs e)
        {
            Point xLocation = new Point(e.X, e.Y);

            if (e.Button == MouseButtons.Left)
            {
                if (xLocation.X < Left)     xLocation.X = Left;
                if (xLocation.X > Right)    xLocation.X = Right;
                if (xLocation.Y < Top)      xLocation.Y = Top;
                if (xLocation.Y > Bottom)   xLocation.Y = Bottom;

                if (m_eControlMode == ControlType.MapCamera)
                {
                    float fAspectRatio = (float)Width / (float)Height;
                    Vector3 xDirection = m_xCamera.Target - m_xCamera.Position;
                    float fDistance = Math.Abs(xDirection.Length());
                    float fHeight = 2.0f * fDistance * (float)Math.Tan(Math.PI / 8.0f);
                    float fPixelSize = fHeight / Height;

                    if (m_xCamera != null)
                    {
                        m_xCamera.MoveRight((xLocation.X - m_xLastMousePosition.X) * -fPixelSize);
                        m_xCamera.MoveUpward((xLocation.Y - m_xLastMousePosition.Y) * fPixelSize);
                    }
                }
                else
                {
                    if (m_xArcBall != null)
                    {
                        m_xArcBall.OnMove(
                            xLocation.X - Left,
                            xLocation.Y - Top,
                            (m_eControlMode == ControlType.RotateObject) ?
                                m_xCamera.Rotation :
                                Quaternion.Identity
                        );
                    }
                }
            }

            m_xLastMousePosition = xLocation;
            UpdateTextOutputControl();
        }

        #endregion

        #region Device States

        public override bool Initialize()
        {
            if (!base.Initialize())
            {
                return false;
            }

            SetupRenderStates();
            SetupLighting();
            SetupCamera();
            SetupControlStates();

            m_xSystem = new ModelExSystem(ref m_xDevice);

            return true;
        }

        protected override void OnResetDevice(object sender, EventArgs e)
        {
            base.OnResetDevice(sender, e);
            SetupLighting();
            Update();

            // May need to reload the mesh with MeshFlags.SystemMemory
            // May need to reload the textures
            // MessageBox.Show("Where is my device?!");
        }

        private void SetupRenderStates()
        {
            CullMode = CullType.None;
            Lighting = true;
            Textures = true;
            TextureFiltering = true;
            Wireframe = false;
            Scaling = true;
        }

        private void SetupLighting()
        {
            m_xDevice.RenderState.Lighting = true;
            m_xDevice.RenderState.Ambient = Color.FromArgb(unchecked((int)0xFF808080));
            m_xDevice.Lights[0].Type = LightType.Directional;
            m_xDevice.Lights[0].Diffuse = Color.FromArgb(unchecked((int)0xFFFFFFFF));
            m_xDevice.Lights[0].Direction = new Vector3(0.0f, -0.3f, 0.5f);
            m_xDevice.Lights[0].Direction.Normalize();
            m_xDevice.Lights[0].Enabled = false;
            m_xDevice.Lights[0].Enabled = true;
            m_xDevice.Lights[0].Position = new Vector3(0.0f, 0.0f, -10000.0f);
        }

        private void SetupCamera()
        {
            m_xCamera = new ModelExCamera(
                ref m_xDevice,
                new Vector3(0.0f, 0.0f, 0.0f),
                2.5f
            );

            UpdateControlStates();
        }

        private void SetupControlStates()
        {
            m_eControlMode = ControlType.RotateObject;
        }

        private void UpdateControlStates()
        {
            switch (m_eControlMode)
            {
                case ControlType.OrbitCamera:
                    if (m_xCamera == null)
                    {
                        goto default;
                    }
                    m_xCamera.Mode = ModelExCamera.CameraMode.Orbit;
                    m_xCamera.Target = new Vector3(0.0f, 0.0f, 0.0f);
                    m_xArcBall = m_xCamera.ArcBall;
                    break;
                case ControlType.FlyCamera:
                    if (m_xCamera == null)
                    {
                        goto default;
                    }
                    m_xCamera.Mode = ModelExCamera.CameraMode.Fly;
                    m_xArcBall = m_xCamera.ArcBall;
                    break;
                case ControlType.MapCamera:
                    if (m_xCamera == null)
                    {
                        goto default;
                    }
                    m_xCamera.Mode = ModelExCamera.CameraMode.Map;
                    m_xArcBall = m_xCamera.ArcBall;
                    break;
                case ControlType.RotateObject:
                    if (m_xObject == null)
                    {
                        goto default;
                    }
                    m_xCamera.Mode = ModelExCamera.CameraMode.Orbit;
                    m_xCamera.Target = new Vector3(0.0f, 0.0f, 0.0f);
                    m_xArcBall = m_xObject.ArcBall;
                    break;
                default:
                    m_xCamera.Mode = ModelExCamera.CameraMode.Manual;
                    m_xArcBall = null;
                    break;
            }
        }

        #endregion

        #region View Modes

        public void ResetRenderStates()
        {
            SetupRenderStates();
        }

        public void ResetCameraPosition()
        {
            SetupCamera();
        }

        public void ResetObjectPosition()
        {
            if (m_xObject != null)
            {
                m_xObject.ArcBall.Rotation = Quaternion.Identity;
            }
        }

        public CullType CullMode
        {
            get { return m_eCullMode; }
            set { m_eCullMode = value; }
        }

        public bool Lighting
        {
            get { return ((m_eRenderFlags & RenderFlags.Lighting) != 0 ); }
            set { if (Lighting != value) m_eRenderFlags ^= RenderFlags.Lighting; }
        }

        public bool Textures
        {
            get { return ((m_eRenderFlags & RenderFlags.Textures) != 0); }
            set { if (Textures != value) m_eRenderFlags ^= RenderFlags.Textures; }
        }

        public bool TextureFiltering
        {
            get { return ((m_eRenderFlags & RenderFlags.TextureFiltering) != 0); }
            set { if (TextureFiltering != value) m_eRenderFlags ^= RenderFlags.TextureFiltering; }
        }

        public bool Wireframe
        {
            get { return ((m_eRenderFlags & RenderFlags.WireFrame) != 0); }
            set { if (Wireframe != value) m_eRenderFlags ^= RenderFlags.WireFrame; }
        }

        public bool Scaling
        {
            get { return ((m_eRenderFlags & RenderFlags.Scaling) != 0); }
            set { if (Scaling != value) m_eRenderFlags ^= RenderFlags.Scaling; }
        }

        public int  ModelIndex
        {
            get { return (m_xObject != null) ? m_xObject.ModelIndex : 0; }
            set { if (m_xObject != null) { m_xObject.ModelIndex = value; } }
        }

        public ModelExObjectInstance Object
        {
            get { return m_xObject; }
        }

        #endregion

        #region Control Modes

        public void ResetControlStates()
        {
            SetupControlStates();
        }

        public ControlType ControlMode
        {
            get { return m_eControlMode; }
            set { m_eControlMode = value; }
        }

        #endregion

        #region Data Output

        public String   TextureArchives
        {
            get { return (m_xObject != null) ? m_xObject.TextureArchives : null; }
        }

        public String   TextureArchiveExt
        {
            get { return (m_xObject != null) ? m_xObject.TextureArchiveExt : null; }
        }

        public String   FileName
        {
            get { return (m_xObject != null) ? m_xObject.FileName : null; }
        }

        public String[] ModelNames
        {
            get
            {
                if (m_xObject != null && m_xObject.Models != null)
                {
                    String[] axModels = new String[m_xObject.Models.Count];
                    for (int m = 0; m < m_xObject.Models.Count; m++)
                    {
                        axModels[m] = m_xObject.Models[m].FileName;
                    }
                    return axModels;
                }
                else
                {
                    return null;
                }
            }
        }

        public String[] TextureNames
        {
            get
            {
                if (m_xObject != null && m_xObject.Textures != null)
                {
                    String[] axTextures = new String[m_xObject.Textures.Count];
                    for (int t = 0; t < m_xObject.Textures.Count; t++)
                    {
                        axTextures[t] = m_xObject.Textures[t].Name;
                    }
                    return axTextures;
                }
                else
                {
                    return null;
                }
            }
        }

        public String[] EffectNames
        {
            get
            {
                if (m_xObject != null && m_xObject.Effects != null)
                {
                    String[] axEffects = new String[m_xObject.Effects.Count];
                    for (int e = 0; e < m_xObject.Effects.Count; e++)
                    {
                        axEffects[e] = m_xObject.Effects[e].Name;
                    }
                    return axEffects;
                }
                else
                {
                    return null;
                }
            }
        }

        public Control  TextOutputControl
        {
            get { return m_xTextOutputControl;}
            set { m_xTextOutputControl = value; }
        }

        protected void UpdateTextOutputControl()
        {
            if (m_xTextOutputControl != null)
            {
                m_xTextOutputControl.Text = "";

                m_xTextOutputControl.Text += "App has focus = ";
                m_xTextOutputControl.Text += (TopLevelControl.ContainsFocus == true) ? "true" : "false";
                m_xTextOutputControl.Text += "\r\n";

                m_xTextOutputControl.Text += "Render view has focus = ";
                m_xTextOutputControl.Text += (Focused == true) ? "true" : "false";
                m_xTextOutputControl.Text += "\r\n";

                m_xTextOutputControl.Text += "Last device reset: " + LastDeviceReset;
                m_xTextOutputControl.Text += "\r\n";

                m_xTextOutputControl.Text += "Last mouse position: " + m_xLastMousePosition.ToString();
                m_xTextOutputControl.Text += "\r\n";
                m_xTextOutputControl.Text += "Last click position: " + m_xLastMouseDownPosition.ToString();
                m_xTextOutputControl.Text += "\r\n";
                m_xTextOutputControl.Text += "Frame rate: " + FrameRate + " fps";
                m_xTextOutputControl.Text += "\r\n";
                m_xTextOutputControl.Text += "Time of last frame: " + LastFrameTime;
                m_xTextOutputControl.Text += "\r\n";

                if (m_xArcBall != null)
                {
                    m_xTextOutputControl.Text += "Dragging arcball = ";
                    m_xTextOutputControl.Text += (m_xArcBall.IsMoving == true) ? "true" : "false";
                    m_xTextOutputControl.Text += "\r\n";
                }

                if (m_xCamera != null)
                {
                    m_xTextOutputControl.Text += String.Format(
                        "Camera Rotation: {{ W = {0:F2}, X = {0:F2}, Y = {1:F2}, Z = {1:F2} }}\r\n",
                        m_xCamera.Rotation.W,
                        m_xCamera.Rotation.X,
                        m_xCamera.Rotation.Y,
                        m_xCamera.Rotation.Z
                    );
                }

                if (m_xObject != null)
                {
                    m_xTextOutputControl.Text += "Textures:";

                    ModelExTextureList axTextures = m_xObject.Textures;
                    if (axTextures != null)
                    {
                        for (int i = 0; i < axTextures.Count; i++)
                        {
                            m_xTextOutputControl.Text += "\r\n\t";
                            String strFilePath = axTextures[i].FileName;
                            if (strFilePath == null || strFilePath == "")
                            {
                                strFilePath = "(No texture)";
                            }
                            m_xTextOutputControl.Text += strFilePath;
                        }
                        m_xTextOutputControl.Text += "\r\n\r\n";
                    }

                    m_xTextOutputControl.Text += "Effects:";

                    ModelExEffectList axEffects = m_xObject.Effects;
                    if (axEffects != null)
                    {
                        for (int i = 0; i < axEffects.Count; i++)
                        {
                            m_xTextOutputControl.Text += "\r\n\t";
                            String strFilePath = axEffects[i].Name;
                            if (strFilePath == null || strFilePath == "")
                            {
                                strFilePath = "(No texture)";
                            }
                            m_xTextOutputControl.Text += strFilePath;
                        }
                        m_xTextOutputControl.Text += "\r\n\r\n";
                    }
                }

                if (m_strLastError != null)
                {
                    m_xTextOutputControl.Text += m_strLastError;
                }
            }
        }

        #endregion

        #region Loading and Saving

        public void LoadModel(String strFileName)
        {
            m_xObject = null;

            try
            {
                m_xSystem.ClearObjectFiles();
                m_xSystem.ClearModelFiles();
                m_xSystem.ClearEffectFiles();
                m_xSystem.ClearTextureFiles();
                GC.Collect();
                ModelExObjectFile xObjectFile = ModelExObjectFile.Create(m_xSystem, strFileName);

                // m_xObject = new ModelExObjectInstance(m_xSystem, strFileName);

                if (m_xObject == null)
                {
                    m_xObject = new ModelExSRObjectInstance(m_xSystem, strFileName);
                    m_xObject.Load(strFileName);
                }

                Control xControl = this;
                m_xObject.ArcBall = new ModelExArcBall(ref xControl, false);

                //ModelExObjectFile xCharacterFile = ModelExObjectFile.Create(m_xSystem, Path.GetDirectoryName(strFileName) + "\\raziel__.drm");
                //m_xCharacter = new ModelExSRObjectInstance(m_xSystem, Path.GetDirectoryName(strFileName) + "\\raziel__.drm");
                //m_xCharacter.Load(Path.GetDirectoryName(strFileName) + "\\raziel__.drm");

                m_strLastError = null;
            }
            catch (Exception xError)
            {
                m_strLastError = xError.Message + "\r\n\r\n" + xError.StackTrace;
                m_xObject = null;
            }

            if (m_xObject != null)
            {
                UpdateControlStates();
                UpdateTextOutputControl();
            }
        }

        public void LoadTexture(String strFileName, int iIndex)
        {
            //if (m_xObject != null &&
            //    iIndex >= 0 && iIndex < m_xObject.Textures.Count)
            //{
            //    m_xObject.Textures.List[iIndex].Load(strFileName);
            //}
        }

        public void LoadTexturesFromFile(String strFileName)
        {
            if (m_xObject != null)
            {
                m_xObject.LoadTexturesFromFile(strFileName);
            }
        }

        public void LoadTexturesFromFolder(String strFolderName)
        {
            if (m_xObject != null)
            {
                m_xObject.LoadTexturesFromFolder(strFolderName);
            }
        }

        #endregion

        #region Updates

        protected override void OnTick()
        {
            base.OnTick();
            UpdateTextOutputControl();
        }

        protected override void RenderScene()
        {
            base.RenderScene();

            if (m_xObject == null || m_xCamera == null)
            {
                return;
            }

            UpdateControlStates();
            m_xCamera.Update();
            m_xDevice.BeginScene();

            float fScale = 0.335799217f;
            if (m_eControlMode == ControlType.MapCamera)
            {
                fScale = 0.06109793f;
            }
            fScale *= 1.0f;
            m_xObject.Render(m_eRenderFlags, m_eCullMode, fScale);
            //m_xCharacter.Render(m_eRenderFlags, m_eCullMode, fScale);
            m_xDevice.EndScene();
        }

        #endregion
    }
}
