using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace ModelEx3
{
    public class RenderPanel : Panel
    {
        protected   Color               m_xBackGroundColour = Color.Blue;
        protected   Device              m_xDevice;
        protected   PresentParameters   m_xPresentParams;

        public RenderPanel()
        {
            SetStyle(ControlStyles.AllPaintingInWmPaint | ControlStyles.Opaque, true);
            Paint += new PaintEventHandler(Render);
        }

        public virtual bool InitializeGraphics()
        {
            return SetupDevice();
        }

        protected virtual bool SetupDevice()
        {
            try
            {
                m_xPresentParams = new PresentParameters();
                m_xPresentParams.Windowed = true;
                m_xPresentParams.SwapEffect = SwapEffect.Discard;
                m_xPresentParams.EnableAutoDepthStencil = true;
                m_xPresentParams.AutoDepthStencilFormat = DepthFormat.D16;
                m_xPresentParams.DeviceWindow = this;

                m_xDevice = new Device(
                    0,
                    DeviceType.Hardware,
                    this.TopLevelControl,
                    CreateFlags.SoftwareVertexProcessing |
                    CreateFlags.MultiThreaded,
                    m_xPresentParams
                );

                m_xDevice.DeviceReset += new EventHandler(OnResetDevice);
            }
            catch (DirectXException)
            {
                return false;
            }

            return true;
        }

        protected virtual void OnResetDevice(object sender, EventArgs e)
        {
        }

        private void Render(object sender, PaintEventArgs e)
        {
            if (m_xDevice == null ||
                TopLevelControl == null ||
                (((Form)TopLevelControl).WindowState == FormWindowState.Minimized))
            {
                return;
            }

            TopLevelControl.Update();

            m_xDevice.Clear(ClearFlags.Target | ClearFlags.ZBuffer, m_xBackGroundColour, 1.0f, 0);

            RenderScene();

            m_xDevice.Present();

            Invalidate();
        }

        protected virtual void RenderScene()
        {
        }
    }
}